import glob
import os
from typing import List, NewType

FilePointer = NewType("FilePointer", str)
"""Unified type for references to files."""


def get_file_pointer_from_path(path: str) -> FilePointer:
    """Returns a file pointer from a path string"""
    return FilePointer(path)


def append_paths_to_pointer(pointer: FilePointer, *paths: str) -> FilePointer:
    """Append directories and/or a file name to a specified file pointer.

    Args:
        pointer: `FilePointer` object to add path to
        paths: any number of directory names optionally followed by a file name to append to the
            pointer

    Returns:
        New file pointer to path given by joining given pointer and path names
    """
    return FilePointer(os.path.join(pointer, *paths))


def does_file_or_directory_exist(pointer: FilePointer) -> bool:
    """Checks if a file or directory exists for a given file pointer

    Args:
        pointer: File Pointer to check if file or directory exists at

    Returns:
        True if file or directory at `pointer` exists, False if not
    """
    return os.path.exists(pointer)


def is_regular_file(pointer: FilePointer) -> bool:
    """Checks if a regular file (NOT a directory) exists for a given file pointer.

    Args:
        pointer: File Pointer to check if a regular file

    Returns:
        True if regular file at `pointer` exists, False if not or is a directory
    """
    return os.path.isfile(pointer)


def find_files_matching_path(*paths: str) -> List[FilePointer]:
    """Find files or directories matching the provided path parts.

    Args:
        paths: any number of directory names optionally followed by a file name.
            directory or file names may be replaced with `*` as a matcher.
    Returns:
        New file pointers to files found matching the path
    """
    matcher = append_paths_to_pointer(*paths)
    return [get_file_pointer_from_path(x) for x in glob.glob(matcher)]


def directory_has_contents(pointer: FilePointer) -> bool:
    """Checks if a directory already has some contents (any files or subdirectories)

    Args:
        pointer: File Pointer to check for existing contents

    Returns:
        True if there are any files or subdirectories below this directory.
    """
    return len(find_files_matching_path(pointer, "*")) > 0


def get_directory_contents(pointer: FilePointer) -> List[FilePointer]:
    """Finds all files and directories in the specified directory.

    NBL This is not recursive, and will return only the first level of directory contents.

    Args:
        pointer: File Pointer in which to find contents

    Returns:
        New file pointers to files or subdirectories below this directory.
    """
    contents = os.listdir(pointer)
    if len(contents) == 0:
        return []
    contents.sort()
    return [append_paths_to_pointer(pointer, x) for x in contents]



"""Tests of catalog functionality"""

import os

import pandas.testing
import pytest

from hipscat.catalog import Catalog, CatalogType, PartitionInfo
from hipscat.pixel_tree.pixel_node_type import PixelNodeType


def test_catalog_load(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    assert len(catalog.get_pixels()) == catalog_pixels.shape[0]
    assert catalog.catalog_name == catalog_info.catalog_name
    for _, pixel in catalog_pixels.iterrows():
        order = pixel[PartitionInfo.METADATA_ORDER_COLUMN_NAME]
        pixel = pixel[PartitionInfo.METADATA_PIXEL_COLUMN_NAME]
        assert (order, pixel) in catalog.pixel_tree
        assert catalog.pixel_tree[(order, pixel)].node_type == PixelNodeType.LEAF


def test_catalog_load_wrong_catalog_info(base_catalog_info, catalog_pixels):
    with pytest.raises(TypeError):
        Catalog(base_catalog_info, catalog_pixels)


def test_catalog_wrong_catalog_type(catalog_info, catalog_pixels):
    catalog_info.catalog_type = CatalogType.INDEX
    with pytest.raises(ValueError):
        Catalog(catalog_info, catalog_pixels)


def test_different_pixel_input_types(catalog_info, catalog_pixels):
    partition_info = PartitionInfo(catalog_pixels)
    catalog = Catalog(catalog_info, partition_info)
    assert len(catalog.get_pixels()) == catalog_pixels.shape[0]
    for _, pixel in catalog_pixels.iterrows():
        order = pixel[PartitionInfo.METADATA_ORDER_COLUMN_NAME]
        pixel = pixel[PartitionInfo.METADATA_PIXEL_COLUMN_NAME]
        assert (order, pixel) in catalog.pixel_tree
        assert catalog.pixel_tree[(order, pixel)].node_type == PixelNodeType.LEAF


def test_wrong_pixel_input_type(catalog_info):
    with pytest.raises(TypeError):
        Catalog(catalog_info, "test")
    with pytest.raises(TypeError):
        Catalog._get_pixel_tree_from_pixels("test")
    with pytest.raises(TypeError):
        Catalog._get_partition_info_from_pixels("test")


def test_get_pixels(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    pixels = catalog.get_pixels()
    pandas.testing.assert_frame_equal(pixels, catalog_pixels)


def test_load_catalog_small_sky(small_sky_dir):
    """Instantiate a catalog with 1 pixel"""
    cat = Catalog.read_from_hipscat(small_sky_dir)

    assert cat.catalog_name == "small_sky"
    assert len(cat.get_pixels()) == 1


def test_load_catalog_small_sky_order1(small_sky_order1_dir):
    """Instantiate a catalog with 4 pixels"""
    cat = Catalog.read_from_hipscat(small_sky_order1_dir)

    assert cat.catalog_name == "small_sky_order1"
    assert len(cat.get_pixels()) == 4


def test_empty_directory(tmp_path):
    """Test loading empty or incomplete data"""
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(os.path.join("path", "empty"))

    catalog_path = os.path.join(tmp_path, "empty")
    os.makedirs(catalog_path, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(catalog_path)

    ## catalog_info file exists - getting closer
    file_name = os.path.join(catalog_path, "catalog_info.json")
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        metadata_file.write('{"catalog_name":"empty", "catalog_type":"source"}')

    with pytest.raises(FileNotFoundError):
        Catalog.read_from_hipscat(catalog_path)

    ## partition_info file exists - enough to create a catalog
    file_name = os.path.join(catalog_path, "partition_info.csv")
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        metadata_file.write("foo")

    catalog = Catalog.read_from_hipscat(catalog_path)
    assert catalog.catalog_name == "empty"



import dataclasses
import os
import os.path

import pandas as pd
import pytest

from hipscat.catalog import PartitionInfo
from hipscat.catalog.association_catalog.association_catalog_info import AssociationCatalogInfo
from hipscat.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.dataset.base_catalog_info import BaseCatalogInfo
from hipscat.inspection.almanac import Almanac

DATA_DIR_NAME = "data"
ALMANAC_DIR_NAME = "almanac"
SMALL_SKY_DIR_NAME = "small_sky"
SMALL_SKY_ORDER1_DIR_NAME = "small_sky_order1"
SMALL_SKY_TO_SMALL_SKY_ORDER1_DIR_NAME = "small_sky_to_small_sky_order1"
TEST_DIR = os.path.dirname(__file__)

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def test_data_dir():
    return os.path.join(TEST_DIR, DATA_DIR_NAME)


@pytest.fixture
def almanac_dir(test_data_dir):
    return os.path.join(test_data_dir, ALMANAC_DIR_NAME)


@pytest.fixture
def small_sky_dir(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_DIR_NAME)


@pytest.fixture
def small_sky_order1_dir(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_ORDER1_DIR_NAME)


@pytest.fixture
def small_sky_to_small_sky_order1_dir(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_TO_SMALL_SKY_ORDER1_DIR_NAME)


@pytest.fixture
def assert_catalog_info_matches_dict():
    def assert_match(catalog_info: BaseCatalogInfo, dictionary: dict):
        """Check that all members of the catalog_info object match dictionary
        elements, where specified."""
        catalog_info_dict = dataclasses.asdict(catalog_info)
        for key, value in dictionary.items():
            assert catalog_info_dict[key] == value

    return assert_match


@pytest.fixture
def base_catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "object",
        "total_rows": 10,
    }


@pytest.fixture
def catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "object",
        "total_rows": 10,
        "epoch": "J2000",
        "ra_column": "ra",
        "dec_column": "dec",
    }


@pytest.fixture
def association_catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "association",
        "total_rows": 10,
        "primary_catalog": "small_sky",
        "primary_column": "id",
        "join_catalog": "small_sky_order1",
        "join_column": "id",
    }


@pytest.fixture
def source_catalog_info() -> dict:
    return {
        "catalog_name": "test_source",
        "catalog_type": "source",
        "total_rows": 100,
        "epoch": "J2000",
        "ra_column": "source_ra",
        "dec_column": "source_dec",
    }


@pytest.fixture
def source_catalog_info_with_extra() -> dict:
    return {
        "catalog_name": "test_source",
        "catalog_type": "source",
        "total_rows": 100,
        "epoch": "J2000",
        "ra_column": "source_ra",
        "dec_column": "source_dec",
        "primary_catalog": "test_name",
        "mjd_column": "mjd",
        "band_column": "band",
        "mag_column": "mag",
        "mag_err_column": "",
    }


@pytest.fixture
def margin_cache_catalog_info() -> dict:
    return {
        "catalog_name": "test_margin",
        "catalog_type": "margin",
        "total_rows": 100,
        "primary_catalog": "test_name",
        "margin_threshold": 0.5,
    }


@pytest.fixture
def index_catalog_info() -> dict:
    return {
        "catalog_name": "test_index",
        "catalog_type": "index",
        "total_rows": 100,
        "primary_catalog": "test_name",
        "indexing_column": "id",
    }


@pytest.fixture
def index_catalog_info_with_extra() -> dict:
    return {
        "catalog_name": "test_index",
        "catalog_type": "index",
        "total_rows": 100,
        "primary_catalog": "test_name",
        "indexing_column": "id",
        "extra_columns": ["foo", "bar"],
    }


@pytest.fixture
def dataset_path(test_data_dir) -> str:
    return os.path.join(test_data_dir, "dataset")


@pytest.fixture
def base_catalog_info_file(dataset_path) -> str:
    return os.path.join(dataset_path, "catalog_info.json")


@pytest.fixture
def base_catalog_info(base_catalog_info_data) -> BaseCatalogInfo:
    return BaseCatalogInfo(**base_catalog_info_data)


@pytest.fixture
def catalog_path(test_data_dir) -> str:
    return os.path.join(test_data_dir, "catalog")


@pytest.fixture
def catalog_info_file(catalog_path) -> str:
    return os.path.join(catalog_path, "catalog_info.json")


@pytest.fixture
def catalog_info(catalog_info_data) -> CatalogInfo:
    return CatalogInfo(**catalog_info_data)


@pytest.fixture
def catalog_pixels() -> pd.DataFrame:
    return pd.DataFrame.from_dict(
        {
            PartitionInfo.METADATA_ORDER_COLUMN_NAME: [1, 1, 2],
            PartitionInfo.METADATA_DIR_COLUMN_NAME: [0, 0, 0],
            PartitionInfo.METADATA_PIXEL_COLUMN_NAME: [0, 1, 8],
        }
    )


@pytest.fixture
def association_catalog_path(test_data_dir) -> str:
    return os.path.join(test_data_dir, "small_sky_to_small_sky_order1")


@pytest.fixture
def association_catalog_info_file(association_catalog_path) -> str:
    return os.path.join(association_catalog_path, "catalog_info.json")


@pytest.fixture
def index_catalog_info_file(test_data_dir) -> str:
    return os.path.join(test_data_dir, "index_catalog", "catalog_info.json")


@pytest.fixture
def margin_cache_catalog_info_file(test_data_dir) -> str:
    return os.path.join(test_data_dir, "margin_cache", "catalog_info.json")


@pytest.fixture
def source_catalog_info_file(test_data_dir) -> str:
    return os.path.join(test_data_dir, "small_sky_source", "catalog_info.json")


@pytest.fixture
def association_catalog_info(association_catalog_info_data) -> AssociationCatalogInfo:
    return AssociationCatalogInfo(**association_catalog_info_data)


@pytest.fixture
def association_catalog_partition_join_file(association_catalog_path) -> str:
    return os.path.join(association_catalog_path, "partition_join_info.csv")


@pytest.fixture
def association_catalog_join_pixels() -> pd.DataFrame:
    return pd.DataFrame.from_dict(
        {
            PartitionJoinInfo.PRIMARY_ORDER_COLUMN_NAME: [0, 0, 0, 0],
            PartitionJoinInfo.PRIMARY_PIXEL_COLUMN_NAME: [11, 11, 11, 11],
            PartitionJoinInfo.JOIN_ORDER_COLUMN_NAME: [1, 1, 1, 1],
            PartitionJoinInfo.JOIN_PIXEL_COLUMN_NAME: [44, 45, 46, 47],
        }
    )


@pytest.fixture
def default_almanac(almanac_dir, test_data_dir):
    """Set up default environment variables and fetch default almanac data."""
    os.environ["HIPSCAT_ALMANAC_DIR"] = almanac_dir
    os.environ["HIPSCAT_DEFAULT_DIR"] = test_data_dir

    return Almanac()



"""Tests of margin bounding utility functions"""

import numpy as np
import numpy.testing as npt
import pytest
from astropy.coordinates import SkyCoord
from regions import PixCoord

import hipscat.pixel_math as pm


def test_get_margin_scale():
    """Check to make sure that get_margin_scale works as expected."""
    scale = pm.get_margin_scale(3, 0.1)

    expected = 1.0272887122119063

    assert scale == expected


def test_get_margin_scale_k_zero():
    """Make sure get_margin_scale works when k == 0"""
    scale = pm.get_margin_scale(0, 0.1)

    expected = 1.0034110890264882

    assert scale == expected


def test_get_margin_scale_k_high():
    """Make sure get_margin_scale works when k is a high order"""
    scale = pm.get_margin_scale(64, 0.1)

    expected = 6.292348628426864e16

    assert scale == expected


def test_negative_margin_threshold():
    """Make sure that get_marin_scale raises a value error when threshold is < 0.0"""
    with pytest.raises(ValueError) as value_error:
        pm.get_margin_scale(3, -0.1)

    assert str(value_error.value) == "margin_threshold must be greater than 0."


def test_zero_margin_threshold():
    """Make sure that get_marin_scale raises a value error when threshold is == 0.0"""
    with pytest.raises(ValueError) as value_error:
        pm.get_margin_scale(3, 0.0)

    assert str(value_error.value) == "margin_threshold must be greater than 0."


def test_get_margin_bounds_and_wcs():
    """Make sure get_margin_bounds_and wcs works as expected."""
    # far out point, expect False
    test_ra1 = 42.0
    test_dec1 = 1.0

    # middle of our healpixel, expect True
    test_ra2 = 56.25
    test_dec2 = 14.49584495

    # actual point from Norder3/Npix4 of my
    # gaia test catalog, expect True
    # if k = 3 and pix = 4
    test_ra3 = 50.61225197
    test_dec3 = 14.4767556

    test_ra = np.array([test_ra1, test_ra2, test_ra3])
    test_dec = np.array([test_dec1, test_dec2, test_dec3])

    test_sc = SkyCoord(test_ra, test_dec, unit="deg")

    scale = pm.get_margin_scale(3, 0.1)
    polygon, wcs_margin = pm.get_margin_bounds_and_wcs(3, 4, scale)[0]

    assert polygon.area == pytest.approx(756822775.0000424, 0.001)

    coord_x, coord_y = wcs_margin.world_to_pixel(test_sc)
    expected_pixel_coord = PixCoord(coord_x, coord_y)
    expected_contains_checks = np.array([False, True, True])

    contains_checks = polygon.contains(expected_pixel_coord)

    npt.assert_array_equal(contains_checks, expected_contains_checks)


def get_checks_for_bounds(bounds, test_sky_coords):
    """Helper method for getting the check flags for
    some boundaries and sky coordinates to check."""
    vals = []
    for polygon, margin in bounds:
        coord_x, coord_y = margin.world_to_pixel(test_sky_coords)

        vals.append(polygon.contains(PixCoord(coord_x, coord_y)))

    return np.array(vals).any(axis=0)


def test_get_margin_bounds_and_wcs_low_order():
    """Make sure get_margin_bounds_and wcs works as expected."""
    # far out point, expect False
    test_ra1 = 90.0
    test_dec1 = -2.0

    # middle of our healpixel, expect True
    test_ra2 = -100.0
    test_dec2 = -45.0

    test_ra = np.array([test_ra1, test_ra2])
    test_dec = np.array([test_dec1, test_dec2])

    test_sc = SkyCoord(test_ra, test_dec, unit="deg")

    scale = pm.get_margin_scale(0, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(0, 5, scale)

    assert len(bounds) == 16

    expected = np.array([True, False])

    npt.assert_array_equal(get_checks_for_bounds(bounds, test_sc), expected)


def test_get_margin_bounds_and_wcs_north_pole():
    """Make sure get_margin_bounds_and_wcs works at pixels along the north polar region"""
    scale = pm.get_margin_scale(1, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(1, 7, scale, step=100)

    assert len(bounds) == 4

    test_ra = np.array([50, 100.0, 130.0, 130.0, 100.0])
    test_dec = np.array([-60.0, 89.9, 65.0, 85.0, 50.0])

    test_sc = SkyCoord(test_ra, test_dec, unit="deg")

    expected = np.array([False, True, True, True, False])

    npt.assert_array_equal(get_checks_for_bounds(bounds, test_sc), expected)


def test_get_margin_bounds_and_wcs_south_pole():
    """Make sure get_margin_bounds_and_wcs works at pixels along the south polar region"""
    scale = pm.get_margin_scale(1, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(1, 36, scale)

    assert len(bounds) == 4

    test_ra = np.array([50, 120.0, 108.0, 150.0, 100.0, 104.0])
    test_dec = np.array([-60.0, -70.0, -66.2, -70.0, -90.0, -80])

    test_sc = SkyCoord(test_ra, test_dec, unit="deg")

    expected = np.array([False, True, True, True, False, True])

    npt.assert_array_equal(get_checks_for_bounds(bounds, test_sc), expected)


def test_get_margin_bounds_and_wcs_ra_rollover():
    """Make sure get_margin_bounds_and_wcs works at the rollover point for right ascension"""
    scale = pm.get_margin_scale(1, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(1, 27, scale)

    assert len(bounds) == 4

    test_ra = np.array([180.0, 0.0, -157.5, 120.0, 157.5])
    test_dec = np.array([20.0, 0.0, 19.4, 20.0, 19.4])

    test_sc = SkyCoord(test_ra, test_dec, unit="deg")

    expected = np.array([True, False, True, False, True])

    npt.assert_array_equal(get_checks_for_bounds(bounds, test_sc), expected)


def test_get_margin_bounds_and_wcs_origin():
    """Make sure get_margin_bounds_and_wcs works at the origin of ra and dec."""
    scale = pm.get_margin_scale(0, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(0, 4, scale)

    assert len(bounds) == 16

    test_ra = np.array([180.0, -20.0])
    test_dec = np.array([20.0, -10.0])

    test_sc = SkyCoord(test_ra, test_dec, unit="deg")

    expected = np.array([False, True])

    npt.assert_array_equal(get_checks_for_bounds(bounds, test_sc), expected)


def test_check_margin_bounds():
    """Make sure check_margin_bounds works properly"""
    scale = pm.get_margin_scale(3, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(3, 4, scale)

    test_ra = np.array([42.4704538, 56.25, 50.61225197])
    test_dec = np.array([1.4593925, 14.49584495, 14.4767556])

    checks = pm.check_margin_bounds(test_ra, test_dec, bounds)

    expected = np.array([False, True, True])

    npt.assert_array_equal(checks, expected)


def test_check_margin_bounds_multi_poly():
    """Make sure check_margin_bounds works when poly_and_wcs has multiple entries"""
    scale = pm.get_margin_scale(1, 0.1)
    bounds = pm.get_margin_bounds_and_wcs(1, 4, scale)

    test_ra = np.array([42.4704538, 120.0, 135.0])
    test_dec = np.array([1.4593925, 25.0, 19.92530172])

    checks = pm.check_margin_bounds(test_ra, test_dec, bounds)

    expected = np.array([False, True, True])

    npt.assert_array_equal(checks, expected)


def test_check_polar_margin_bounds_north():
    """Make sure check_polar_margin_bounds works at the north pole"""
    order = 0
    pix = 1
    margin_order = 2

    test_ra = np.array([89, -179, -45])
    test_dec = np.array([89.9, 89.9, 85.0])

    vals = pm.check_polar_margin_bounds(test_ra, test_dec, order, pix, margin_order, 0.1)

    expected = np.array([True, True, False])

    npt.assert_array_equal(vals, expected)


def test_check_polar_margin_bounds_south():
    """Make sure check_polar_margin_bounds works at the south pole"""
    order = 0
    pix = 9
    margin_order = 2

    test_ra = np.array([89, -179, -45])
    test_dec = np.array([-89.9, -89.9, -85.0])

    vals = pm.check_polar_margin_bounds(test_ra, test_dec, order, pix, margin_order, 0.1)

    expected = np.array([True, True, False])

    npt.assert_array_equal(vals, expected)


def test_check_polar_margin_bounds_non_pole():
    """Make sure check_polar_margin_bounds raises a ValueError when pix isn't polar"""
    order = 0
    pix = 7
    margin_order = 2

    test_ra = np.array([89, -179, -45])
    test_dec = np.array([-89.9, -89.9, -85.0])

    with pytest.raises(ValueError) as value_error:
        pm.check_polar_margin_bounds(test_ra, test_dec, order, pix, margin_order, 0.1)

    assert str(value_error.value) == "provided healpixel must be polar"



"""Generate a molleview map with the pixel densities of the catalog

NB: Testing validity of generated plots is currently not tested in our unit test suite.
"""

import healpy as hp
import numpy as np
from matplotlib import pyplot as plt

from hipscat.catalog import Catalog, PartitionInfo
from hipscat.io import file_io, paths


def _read_point_map(catalog_base_dir):
    """Read the object spatial distribution information from a healpix FITS file.

    Args:
        catalog_base_dir: path to a catalog
    Returns:
        one-dimensional numpy array of long integers where the value at each index
        corresponds to the number of objects found at the healpix pixel.
    """
    map_file_pointer = paths.get_point_map_file_pointer(catalog_base_dir)
    return file_io.read_fits_image(map_file_pointer)


def plot_points(catalog: Catalog, projection="moll", draw_map=True):
    """Create a visual map of the input points of the catalog.

    Args:
        catalog (`hipscat.catalog.Catalog`) Catalog to display
        projection (str) The map projection to use. Valid values include:
            - moll - Molleweide projection (default)
            - gnom - Gnomonic projection
            - cart - Cartesian projection
            - orth - Orthographic projection
    """
    point_map = _read_point_map(catalog.catalog_base_dir)
    _plot_healpix_map(
        point_map,
        projection,
        f"Catalog point density map - {catalog.catalog_name}",
        draw_map=draw_map,
    )


def plot_pixels(catalog: Catalog, projection="moll", draw_map=True):
    """Create a visual map of the pixel density of the catalog.

    Args:
        catalog (`hipscat.catalog.Catalog`) Catalog to display
        projection (str) The map projection to use. Valid values include:
            - moll - Molleweide projection (default)
            - gnom - Gnomonic projection
            - cart - Cartesian projection
            - orth - Orthographic projection
    """
    pixels = catalog.get_pixels()

    catalog_orders = pixels[PartitionInfo.METADATA_ORDER_COLUMN_NAME].unique()
    catalog_orders.sort()
    max_order = catalog_orders[-1]

    order_map = np.full(hp.order2npix(max_order), hp.pixelfunc.UNSEEN)

    for _, pixel in pixels.iterrows():
        explosion_factor = 4 ** (max_order - pixel[PartitionInfo.METADATA_ORDER_COLUMN_NAME])
        exploded_pixels = [
            *range(
                pixel[PartitionInfo.METADATA_PIXEL_COLUMN_NAME] * explosion_factor,
                (pixel[PartitionInfo.METADATA_PIXEL_COLUMN_NAME] + 1) * explosion_factor,
            )
        ]
        order_map[exploded_pixels] = pixel[PartitionInfo.METADATA_ORDER_COLUMN_NAME]
    _plot_healpix_map(
        order_map,
        projection,
        f"Catalog pixel density map - {catalog.catalog_name}",
        draw_map=draw_map,
    )


def _plot_healpix_map(healpix_map, projection, title, draw_map=True):
    """Perform the plotting of a healpix pixel map.

    Args:
        healpix_map: array containing the map
        projection: projection type to display
        title: title used in image plot
    """
    if projection == "moll":
        projection_method = hp.mollview
    elif projection == "gnom":
        projection_method = hp.gnomview
    elif projection == "cart":
        projection_method = hp.cartview
    elif projection == "orth":
        projection_method = hp.orthview
    else:
        raise NotImplementedError(f"unknown projection: {projection}")

    if draw_map:  # pragma: no cover
        projection_method(
            healpix_map,
            title=title,
            nest=True,
        )
        plt.plot()



"""Container class to hold per-partition metadata"""

from typing import List

import numpy as np
import pandas as pd

from hipscat.io import FilePointer, file_io
from hipscat.pixel_math import HealpixPixel


class PartitionInfo:
    """Container class for per-partition info."""

    METADATA_ORDER_COLUMN_NAME = "Norder"
    METADATA_DIR_COLUMN_NAME = "Dir"
    METADATA_PIXEL_COLUMN_NAME = "Npix"

    def __init__(self, pixels: pd.DataFrame) -> None:
        self.data_frame = pixels

    def get_healpix_pixels(self) -> List[HealpixPixel]:
        """Get healpix pixel objects for all pixels represented as partitions.

        Returns:
            List of HealpixPixel
        """
        return [
            HealpixPixel(order, pixel)
            for order, pixel in zip(
                self.data_frame[self.METADATA_ORDER_COLUMN_NAME],
                self.data_frame[self.METADATA_PIXEL_COLUMN_NAME],
            )
        ]

    def get_highest_order(self) -> int:
        """Get the highest healpix order for the dataset.

        Returns:
            int representing highest order.
        """
        highest_order = np.max(self.data_frame[self.METADATA_ORDER_COLUMN_NAME].values)

        return highest_order

    @classmethod
    def read_from_file(cls, partition_info_file: FilePointer):
        """Read partition info from a `partition_info.csv` file to create an object

        Args:
            partition_info_file: FilePointer to the `partition_info.csv` file

        Returns:
            A `PartitionInfo` object with the data from the file
        """
        if not file_io.does_file_or_directory_exist(partition_info_file):
            raise FileNotFoundError(f"No partition info found where expected: {str(partition_info_file)}")

        data_frame = file_io.load_csv_to_pandas(partition_info_file)
        return cls(data_frame)



# pylint: disable=duplicate-code

from __future__ import annotations

import pandas as pd

from hipscat.catalog.partition_info import PartitionInfo
from hipscat.pixel_math.healpix_pixel_convertor import HealpixInputTypes, get_healpix_pixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree import PixelTree


class PixelTreeBuilder:
    """Build a PixelTree

    Initially a root node is created when the builder is initialized.
    Nodes can then be added to the tree.
    To create a pixel tree object once the tree is built, call the `build` method

    """

    def __init__(self):
        self.root_pixel = PixelNode((-1, -1), PixelNodeType.ROOT, None)
        self.pixels = {-1: {-1: self.root_pixel}}

    def build(self) -> PixelTree:
        """Build a `PixelTree` object from the nodes created

        Returns:
            The pixel tree with the nodes created in the builder
        """
        return PixelTree(self.root_pixel, self.pixels)

    @staticmethod
    def from_partition_info_df(partition_info_df: pd.DataFrame) -> PixelTree:
        """Build a tree from a partition_info DataFrame

        Args:
            partition_info_df: Pandas Dataframe with columns matching those in the
            `partition_info.csv` metadata file

        Returns:
            The pixel tree with the leaf pixels specified in the DataFrame
        """
        builder = PixelTreeBuilder()
        # pylint: disable=protected-access
        builder._create_tree_from_partition_info_df(partition_info_df)
        return builder.build()

    def contains(self, pixel: HealpixInputTypes) -> bool:
        """Check if tree contains a node at a given order and pixel

        Args:
            pixel: HEALPix pixel to check. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            True if the tree contains the pixel, False if not
        """
        pixel = get_healpix_pixel(pixel)
        return pixel.order in self.pixels and pixel.pixel in self.pixels[pixel.order]

    def __contains__(self, item):
        return self.contains(item)

    def get_node(self, pixel: HealpixInputTypes) -> PixelNode | None:
        """Get the node at a given pixel

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            The PixelNode at the index, or None if a node does not exist
        """
        pixel = get_healpix_pixel(pixel)
        if self.contains(pixel):
            return self.pixels[pixel.order][pixel.pixel]
        return None

    def __getitem__(self, item):
        return self.get_node(item)

    def _create_tree_from_partition_info_df(self, partition_info_df: pd.DataFrame):
        """Creates the tree by recursively creating parent nodes until reaching the root from
        each leaf pixel

        Validates the pixel structure as it does so by checking for duplicate pixels or pixels
        defined at multiple orders

        Args:
            partition_info_df: Dataframe loaded from the partition_info metadata
        """
        for _, row in partition_info_df.iterrows():
            self.create_node_and_parent_if_not_exist(
                (
                    row[PartitionInfo.METADATA_ORDER_COLUMN_NAME],
                    row[PartitionInfo.METADATA_PIXEL_COLUMN_NAME],
                ),
                PixelNodeType.LEAF,
            )

    def create_node_and_parent_if_not_exist(self, pixel: HealpixInputTypes, node_type: PixelNodeType):
        """Creates a node and adds to `self.pixels` in the tree, and recursively creates parent
        node if parent does not exist

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)
            node_type: Node type of the node to create
        """
        pixel = get_healpix_pixel(pixel)
        if self.contains(pixel):
            raise ValueError("Incorrectly configured catalog: catalog contains duplicate pixels")

        if pixel.order == 0:
            self.create_node(pixel, node_type, self.root_pixel)
            return

        parent_order = pixel.order - 1
        parent_pixel = pixel.pixel >> 2
        if not self.contains((parent_order, parent_pixel)):
            self.create_node_and_parent_if_not_exist((parent_order, parent_pixel), PixelNodeType.INNER)

        parent = self.pixels[parent_order][parent_pixel]

        if parent.node_type != PixelNodeType.INNER:
            raise ValueError(
                "Incorrectly configured catalog: catalog contains pixels defined at multiple orders"
            )

        self.create_node(pixel, node_type, parent)

    def create_node(self, pixel: HealpixInputTypes, node_type: PixelNodeType, parent: PixelNode):
        """Create a node and add to `self.pixels` in the tree

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)
            node_type: Node type of the node to create
            parent: Parent node of the node to create
        """
        pixel = get_healpix_pixel(pixel)
        node = PixelNode(pixel, node_type, parent)
        if pixel.order not in self.pixels:
            self.pixels[pixel.order] = {}
        self.pixels[pixel.order][pixel.pixel] = node



import pytest

from hipscat.pixel_math.healpix_pixel import HealpixPixel


def test_pixels_equal():
    order = 3
    pixel = 42
    pix1 = HealpixPixel(order=order, pixel=pixel)
    pix2 = HealpixPixel(order=order, pixel=pixel)
    assert pix1 == pix2
    assert pix1.order == pix2.order
    assert pix1.pixel == pix2.pixel


def test_order_greater_than_max_order_fails():
    with pytest.raises(ValueError):
        HealpixPixel(order=20, pixel=0)


def test_equal_pixel_hash_equal():
    order = 3
    pixel = 42
    test_string = "testing"
    pix1 = HealpixPixel(order=order, pixel=pixel)
    test_dict = {}
    test_dict[pix1] = test_string
    pix2 = HealpixPixel(order=order, pixel=pixel)
    assert pix1 == pix2
    assert pix2 in test_dict
    assert test_dict[pix2] == test_string


def test_pixel_str_and_repr():
    order = 3
    pixel = 42
    test_string = f"Order: {order}, Pixel: {pixel}"
    pix = HealpixPixel(order=order, pixel=pixel)
    assert str(pix) == test_string
    assert repr(pix) == test_string



import pandas as pd
import pytest

from hipscat.catalog import PartitionInfo
from hipscat.pixel_math import HealpixPixel
from hipscat.pixel_tree.pixel_node import PixelNode
from hipscat.pixel_tree.pixel_node_type import PixelNodeType
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


def assert_pixel_tree_has_nodes_in_catalog(tree, catalog):
    """assert tree contains the same nodes as the catalog"""
    assert tree.contains((-1, -1))
    for _, pixel in catalog.get_pixels().iterrows():
        assert tree.contains(
            (
                pixel[PartitionInfo.METADATA_ORDER_COLUMN_NAME],
                pixel[PartitionInfo.METADATA_PIXEL_COLUMN_NAME],
            )
        )


def test_pixel_tree_small_sky(small_sky_catalog, small_sky_pixels):
    """test pixel tree on small sky"""
    pixel_tree = PixelTreeBuilder.from_partition_info_df(small_sky_catalog.get_pixels())
    assert len(pixel_tree) == len(small_sky_catalog.get_pixels()) + 1
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_catalog)
    small_sky_pixel = pixel_tree.get_node(small_sky_pixels[0])
    assert small_sky_pixel.parent == pixel_tree.root_pixel
    assert pixel_tree.root_pixel.node_type == PixelNodeType.ROOT


def test_pixel_tree_small_sky_order1(small_sky_order1_catalog, small_sky_order1_pixels):
    """test pixel tree on small sky order1"""
    pixel_tree = PixelTreeBuilder.from_partition_info_df(small_sky_order1_catalog.get_pixels())
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_order1_catalog)
    first_pixel = pixel_tree.get_node(small_sky_order1_pixels[0])
    second_pixel = pixel_tree.get_node(small_sky_order1_pixels[1])
    parent_node = first_pixel.parent
    assert parent_node.node_type == PixelNodeType.INNER
    assert first_pixel.parent == second_pixel.parent
    assert parent_node.parent == pixel_tree.root_pixel


def test_duplicate_pixel_raises_error(small_sky_catalog):
    """test pixel tree raises error with duplicate pixels"""
    partition_info = small_sky_catalog.get_pixels()
    pixel_row = partition_info.iloc[0]
    info_with_duplicate = pd.concat([partition_info, pixel_row.to_frame().T])
    with pytest.raises(ValueError):
        PixelTreeBuilder.from_partition_info_df(info_with_duplicate)


def test_pixel_duplicated_at_different_order_raises_error(small_sky_catalog):
    """test pixel tree raises error with duplicate pixels at different orders"""
    partition_info = small_sky_catalog.get_pixels()
    pixel_row = partition_info.iloc[0].copy()
    pixel_row[PartitionInfo.METADATA_ORDER_COLUMN_NAME] += 1
    pixel_row[PartitionInfo.METADATA_PIXEL_COLUMN_NAME] = (
        pixel_row[PartitionInfo.METADATA_PIXEL_COLUMN_NAME] << 2
    )
    info_with_duplicate = pd.concat([partition_info, pixel_row.to_frame().T])
    with pytest.raises(ValueError):
        PixelTreeBuilder.from_partition_info_df(info_with_duplicate)


def test_tree_builder_contains_root_node():
    tree_builder = PixelTreeBuilder()
    assert tree_builder.contains(HealpixPixel(-1, -1))
    assert tree_builder.contains((-1, -1))
    assert (-1, -1) in tree_builder
    assert HealpixPixel(-1, -1) in tree_builder
    assert not tree_builder.contains(HealpixPixel(10, 10))
    assert not tree_builder.contains((10, 10))
    assert (10, 10) not in tree_builder
    assert HealpixPixel(10, 10) not in tree_builder


def test_pixel_builder_contains_added_node():
    tree_builder = PixelTreeBuilder()
    assert not tree_builder.contains((0, 0))
    tree_builder.create_node((0, 0), PixelNodeType.LEAF, tree_builder.root_pixel)
    assert tree_builder.contains((0, 0))


def test_pixel_builder_retrieve_added_node():
    tree_builder = PixelTreeBuilder()
    order = 0
    pixel = 0
    node_type = PixelNodeType.LEAF
    tree_builder.create_node((order, pixel), node_type, tree_builder.root_pixel)
    assert isinstance(tree_builder.get_node((order, pixel)), PixelNode)
    assert tree_builder.get_node((order, pixel)).node_type == node_type
    assert tree_builder.get_node((order, pixel)).parent == tree_builder.root_pixel
    assert tree_builder.get_node((order, pixel)).hp_order == order
    assert tree_builder.get_node((order, pixel)).hp_pixel == pixel
    assert tree_builder.get_node(HealpixPixel(order, pixel)) == tree_builder.get_node((order, pixel))
    assert tree_builder[(order, pixel)] == tree_builder.get_node((order, pixel))
    assert tree_builder[HealpixPixel(order, pixel)] == tree_builder.get_node((order, pixel))


def test_pixel_builder_retrieve_none_node():
    tree_builder = PixelTreeBuilder()
    assert tree_builder.get_node((10, 10)) is None
    assert tree_builder.get_node(HealpixPixel(10, 10)) is None
    assert tree_builder[(10, 10)] is None
    assert tree_builder[HealpixPixel(10, 10)] is None


def test_built_pixel_tree_contains_same_nodes_as_builder():
    tree_builder = PixelTreeBuilder()
    tree_builder.create_node((0, 0), PixelNodeType.INNER, tree_builder.root_pixel)
    tree_builder.create_node((1, 1), PixelNodeType.LEAF, tree_builder[(0, 0)])
    tree = tree_builder.build()
    for order, nodes_at_order in tree_builder.pixels.items():
        for pixel, node in nodes_at_order.items():
            assert tree.get_node((order, pixel)) == node
            assert tree_builder.get_node((order, pixel)) == node



"""Container class to hold catalog metadata and partition iteration"""
from __future__ import annotations

from typing import Tuple, Union

import pandas as pd

from hipscat.catalog.catalog_info import CatalogInfo
from hipscat.catalog.catalog_type import CatalogType
from hipscat.catalog.dataset.dataset import Dataset
from hipscat.catalog.partition_info import PartitionInfo
from hipscat.io import FilePointer, file_io, paths
from hipscat.pixel_tree.pixel_tree import PixelTree
from hipscat.pixel_tree.pixel_tree_builder import PixelTreeBuilder


class Catalog(Dataset):
    """A HiPSCat Catalog with data stored in a HEALPix Hive partitioned structure

    Catalogs of this type are partitioned spatially, contain `partition_info` metadata specifying
    the pixels in Catalog, and on disk conform to the parquet partitioning structure
    `Norder=/Dir=/Npix=.parquet`
    """

    CatalogInfoClass = CatalogInfo
    PixelInputTypes = Union[pd.DataFrame, PartitionInfo]
    HIPS_CATALOG_TYPES = [CatalogType.OBJECT, CatalogType.SOURCE, CatalogType.MARGIN]

    def __init__(
        self,
        catalog_info: CatalogInfoClass,
        pixels: PixelInputTypes,
        catalog_path: str = None,
    ) -> None:
        """Initializes a Catalog

        Args:
            catalog_info: CatalogInfo object with catalog metadata
            pixels: Specifies the pixels contained in the catalog. Can be either a Dataframe with
                columns `Norder`, `Dir`, and `Npix` matching a `partition_info.csv` file, or a
                PartitionInfo object
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
        """
        if catalog_info.catalog_type not in self.HIPS_CATALOG_TYPES:
            raise ValueError(
                f"Catalog info `catalog_type` must be one of "
                f"{', '.join([t.value for t in self.HIPS_CATALOG_TYPES])}"
            )
        super().__init__(catalog_info, catalog_path)
        self.partition_info = self._get_partition_info_from_pixels(pixels)
        self.pixel_tree = self._get_pixel_tree_from_pixels(pixels)

    @staticmethod
    def _get_partition_info_from_pixels(pixels: PixelInputTypes) -> PartitionInfo:
        if isinstance(pixels, PartitionInfo):
            return pixels
        if isinstance(pixels, pd.DataFrame):
            return PartitionInfo(pixels)
        raise TypeError("Pixels must be of type PartitionInfo or Dataframe")

    @staticmethod
    def _get_pixel_tree_from_pixels(pixels: PixelInputTypes) -> PixelTree:
        if isinstance(pixels, PartitionInfo):
            return PixelTreeBuilder.from_partition_info_df(pixels.data_frame)
        if isinstance(pixels, pd.DataFrame):
            return PixelTreeBuilder.from_partition_info_df(pixels)
        raise TypeError("Pixels must be of type PartitionInfo or Dataframe")

    def get_pixels(self):
        """Get all healpix pixels that are contained in the catalog

        Returns:
            data frame with per-pixel data.

            The data frame contains the following columns:

            - order: order of the destination pixel
            - pixel: pixel number *at the above order*
            - num_objects: the number of rows in the pixel's partition
        """
        return self.partition_info.data_frame

    @classmethod
    def _read_args(cls, catalog_base_dir: FilePointer) -> Tuple[CatalogInfoClass, PartitionInfo]:
        args = super()._read_args(catalog_base_dir)
        partition_info_file = paths.get_partition_info_pointer(catalog_base_dir)
        partition_info = PartitionInfo.read_from_file(partition_info_file)
        return args + (partition_info,)

    @classmethod
    def _check_files_exist(cls, catalog_base_dir: FilePointer):
        super()._check_files_exist(catalog_base_dir)
        partition_info_file = paths.get_partition_info_pointer(catalog_base_dir)
        if not file_io.does_file_or_directory_exist(partition_info_file):
            raise FileNotFoundError(f"No partition info found where expected: {str(partition_info_file)}")



"""Tests of file IO (reads and writes)"""

import os
import shutil

import numpy.testing as npt
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from hipscat.io import file_io
import hipscat.io.write_metadata as io
import hipscat.pixel_math as hist
from hipscat.pixel_math.healpix_pixel import HealpixPixel


def test_write_json_file(assert_text_file_matches, tmp_path):
    """Test of arbitrary json dictionary with strings and numbers"""

    expected_lines = [
        "{\n",
        '    "first_english": "a",',
        '    "first_greek": "alpha",',
        '    "first_number": 1,',
        r'    "first_five_fib": \[',
        "        1,",
        "        1,",
        "        2,",
        "        3,",
        "        5",
        "    ]",
        "}",
    ]

    dictionary = {}
    dictionary["first_english"] = "a"
    dictionary["first_greek"] = "alpha"
    dictionary["first_number"] = 1
    dictionary["first_five_fib"] = [1, 1, 2, 3, 5]

    json_filename = os.path.join(tmp_path, "dictionary.json")
    io.write_json_file(dictionary, json_filename)
    assert_text_file_matches(expected_lines, json_filename)


def test_write_catalog_info(assert_text_file_matches, tmp_path, catalog_info):
    """Test that we accurately write out catalog metadata"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    os.makedirs(catalog_base_dir)
    expected_lines = [
        "{",
        '    "catalog_name": "test_name",',
        '    "catalog_type": "object",',
        '    "total_rows": 10,',
        '    "epoch": "J2000",',
        '    "ra_column": "ra",',
        '    "dec_column": "dec"',
        "}",
    ]

    io.write_catalog_info(dataset_info=catalog_info, catalog_base_dir=catalog_base_dir)
    metadata_filename = os.path.join(catalog_base_dir, "catalog_info.json")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_provenance_info(assert_text_file_matches, tmp_path, catalog_info):
    """Test that we accurately write out tool-provided generation metadata"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    os.makedirs(catalog_base_dir)
    expected_lines = [
        "{",
        '    "catalog_name": "test_name",',
        '    "catalog_type": "object",',
        '    "total_rows": 10,',
        '    "epoch": "J2000",',
        '    "ra_column": "ra",',
        '    "dec_column": "dec",',
        r'    "version": ".*",',  # version matches digits
        r'    "generation_date": "[.\d]+",',  # date matches date format
        '    "tool_args": {',
        '        "tool_name": "hipscat-import",',
        '        "tool_version": "1.0.0",',
        r'        "input_file_names": \[',
        '            "file1",',
        '            "file2",',
        '            "file3"',
        "        ]",
        "    }",
        "}",
    ]

    tool_args = {
        "tool_name": "hipscat-import",
        "tool_version": "1.0.0",
        "input_file_names": ["file1", "file2", "file3"],
    }

    io.write_provenance_info(
        catalog_base_dir=catalog_base_dir, dataset_info=catalog_info, tool_args=tool_args
    )
    metadata_filename = os.path.join(catalog_base_dir, "provenance_info.json")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_partition_info_healpix_pixel_map(assert_text_file_matches, tmp_path):
    """Test that we accurately write out the partition stats for overloaded input"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    os.makedirs(catalog_base_dir)
    expected_lines = [
        "Norder,Dir,Npix,num_rows",
        "0,0,11,131",
    ]
    pixel_map = {HealpixPixel(0, 11): (131, [11])}
    io.write_partition_info(catalog_base_dir, destination_healpix_pixel_map=pixel_map)
    metadata_filename = os.path.join(catalog_base_dir, "partition_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)

    expected_lines = [
        "Norder,Dir,Npix,num_rows",
        "1,0,44,51",
        "1,0,45,29",
        "1,0,46,51",
    ]
    pixel_map = {
        HealpixPixel(1, 44): (51, [44]),
        HealpixPixel(1, 45): (29, [45]),
        HealpixPixel(1, 46): (51, [46]),
    }
    io.write_partition_info(catalog_base_dir, destination_healpix_pixel_map=pixel_map)
    metadata_filename = os.path.join(catalog_base_dir, "partition_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_partition_info_float(assert_text_file_matches, tmp_path):
    """Test that we accurately write out the individual partition stats
    even when the input is floats instead of ints"""
    catalog_base_dir = os.path.join(tmp_path, "test_name")
    os.makedirs(catalog_base_dir)
    expected_lines = [
        "Norder,Dir,Npix,num_rows",
        "0,0,11,131",
    ]
    pixel_map = {HealpixPixel(0.0, 11.0): (131, [44.0, 45.0, 46.0])}
    io.write_partition_info(catalog_base_dir, pixel_map)
    metadata_filename = os.path.join(catalog_base_dir, "partition_info.csv")
    assert_text_file_matches(expected_lines, metadata_filename)


def test_write_parquet_metadata(tmp_path, small_sky_dir, basic_catalog_parquet_metadata):
    """Copy existing catalog and create new metadata files for it"""
    catalog_base_dir = os.path.join(tmp_path, "catalog")
    shutil.copytree(
        small_sky_dir,
        catalog_base_dir,
    )
    io.write_parquet_metadata(catalog_base_dir)
    check_parquet_schema(os.path.join(catalog_base_dir, "_metadata"), basic_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(catalog_base_dir, "_common_metadata"),
        basic_catalog_parquet_metadata,
        0,
    )
    ## Re-write - should still have the same properties.
    io.write_parquet_metadata(catalog_base_dir)
    check_parquet_schema(os.path.join(catalog_base_dir, "_metadata"), basic_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(catalog_base_dir, "_common_metadata"),
        basic_catalog_parquet_metadata,
        0,
    )


def test_write_parquet_metadata_order1(tmp_path, small_sky_order1_dir, basic_catalog_parquet_metadata):
    """Copy existing catalog and create new metadata files for it,
    using a catalog with multiple files."""
    temp_path = os.path.join(tmp_path, "catalog")
    shutil.copytree(
        small_sky_order1_dir,
        temp_path,
    )

    io.write_parquet_metadata(temp_path)
    ## 4 row groups for 4 partitioned parquet files
    check_parquet_schema(
        os.path.join(temp_path, "_metadata"),
        basic_catalog_parquet_metadata,
        4,
    )
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(temp_path, "_common_metadata"),
        basic_catalog_parquet_metadata,
        0,
    )


def test_write_index_parquet_metadata(tmp_path):
    """Create an index-like catalog, and test metadata creation."""
    temp_path = os.path.join(tmp_path, "index")

    index_parquet_path = os.path.join(temp_path, "Parts=0", "part_000_of_001.parquet")
    os.makedirs(os.path.join(temp_path, "Parts=0"))
    basic_index = pd.DataFrame({"_hipscat_id": [4000, 4001], "ps1_objid": [700, 800]})
    basic_index.to_parquet(index_parquet_path)

    index_catalog_parquet_metadata = pa.schema(
        [
            pa.field("_hipscat_id", pa.int64()),
            pa.field("ps1_objid", pa.int64()),
        ]
    )

    io.write_parquet_metadata(temp_path)
    check_parquet_schema(os.path.join(tmp_path, "index", "_metadata"), index_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        os.path.join(tmp_path, "index", "_common_metadata"),
        index_catalog_parquet_metadata,
        0,
    )


def check_parquet_schema(file_name, expected_schema, expected_num_row_groups=1):
    """Check parquet schema against expectations"""
    assert os.path.exists(file_name), f"file not found [{file_name}]"

    single_metadata = pq.read_metadata(file_name)
    schema = single_metadata.schema.to_arrow_schema()

    assert len(schema) == len(
        expected_schema
    ), f"object list not the same size ({len(schema)} vs {len(expected_schema)})"

    npt.assert_array_equal(schema.names, expected_schema.names)

    assert schema.equals(expected_schema, check_metadata=False)

    parquet_file = pq.ParquetFile(file_name)
    assert parquet_file.metadata.num_row_groups == expected_num_row_groups

    for row_index in range(0, parquet_file.metadata.num_row_groups):
        row_md = parquet_file.metadata.row_group(row_index)
        for column_index in range(0, row_md.num_columns):
            column_metadata = row_md.column(column_index)
            assert column_metadata.file_path.endswith(".parquet")


def test_read_write_fits_point_map(tmp_path):
    """Check that we write and can read a FITS file for spatial distribution."""
    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [51, 29, 51, 0]
    initial_histogram[44:] = filled_pixels[:]
    io.write_fits_map(tmp_path, initial_histogram)

    output_file = os.path.join(tmp_path, "point_map.fits")

    output = file_io.read_fits_image(output_file)
    npt.assert_array_equal(output, initial_histogram)



import json
import os
import shutil
from typing import Any

import healpy as hp
import numpy as np
import pandas as pd
import pyarrow.parquet as pq

from hipscat.io.file_io.file_pointer import FilePointer


def make_directory(file_pointer: FilePointer, exist_ok: bool = False):
    """Make a directory at a given file pointer

    Will raise an error if a directory already exists, unless `exist_ok` is True in which case
    any existing directories will be left unmodified

    Args:
        file_pointer: location in file system to make directory
        exist_ok: Default False. If false will raise error if directory exists. If true existing
            directories will be ignored and not modified

    Raises:
        OSError
    """
    os.makedirs(file_pointer, exist_ok=exist_ok)


def remove_directory(file_pointer: FilePointer, ignore_errors=False):
    """Remove a directory, and all contents, recursively.

    Args:
        file_pointer: directory in file system to remove
        ignore_errors: if True errors resulting from failed removals will be ignored
    """
    shutil.rmtree(file_pointer, ignore_errors=ignore_errors)


def write_string_to_file(file_pointer: FilePointer, string: str, encoding: str = "utf-8"):
    """Write a string to a text file

    Args:
        file_pointer: file location to write file to
        string: string to write to file
        encoding: Default: 'utf-8', encoding method to write to file with
    """
    with open(file_pointer, "w", encoding=encoding) as file:
        file.write(string)


def load_json_file(file_pointer: FilePointer, encoding: str = "utf-8") -> dict:
    """Load a json file to a dictionary

    Args:
        file_pointer: location of file to read
        encoding: string encoding method used by the file
    Returns:
        dictionary of key value pairs loaded from the JSON file
    """
    json_dict = None
    with open(file_pointer, "r", encoding=encoding) as json_file:
        json_dict = json.load(json_file)
    return json_dict


def load_csv_to_pandas(file_pointer: FilePointer, **kwargs) -> pd.DataFrame:
    """Load a csv file to a pandas dataframe

    Args:
        file_pointer: location of csv file to load
        **kwargs: arguments to pass to pandas `read_csv` loading method
    Returns:
        pandas dataframe loaded from CSV
    """
    return pd.read_csv(file_pointer, **kwargs)


def write_dataframe_to_csv(dataframe: pd.DataFrame, file_pointer: FilePointer, **kwargs):
    """Write a pandas DataFrame to a CSV file

    Args:
        dataframe: DataFrame to write
        file_pointer: location of file to write to
        **kwargs: args to pass to pandas `to_csv` method
    """
    dataframe.to_csv(file_pointer, **kwargs)


def read_parquet_metadata(file_pointer: FilePointer, **kwargs) -> pq.FileMetaData:
    """Read FileMetaData from footer of a single Parquet file.

    Args:
        file_pointer: location of file to read metadata from
        **kwargs: additional arguments to be passed to pyarrow.parquet.read_metadata
    """
    return pq.read_metadata(file_pointer, **kwargs)


def write_parquet_metadata(schema: Any, file_pointer: FilePointer, metadata_collector: list = None, **kwargs):
    """Write a metadata only parquet file from a schema

    Args:
        schema: schema to be written
        file_pointer: location of file to be written to
        metadata_collector: where to collect metadata information
        **kwargs: additional arguments to be passed to pyarrow.parquet.write_metadata
    """
    pq.write_metadata(schema, file_pointer, metadata_collector=metadata_collector, **kwargs)


def read_fits_image(map_file_pointer: FilePointer):
    """Read the object spatial distribution information from a healpix FITS file.

    Args:
        file_pointer: location of file to be written
    Returns:
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
    """
    return hp.read_map(map_file_pointer)


def write_fits_image(histogram: np.ndarray, map_file_pointer: FilePointer):
    """Write the object spatial distribution information to a healpix FITS file.

    Args:
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        file_pointer: location of file to be written
    """
    hp.write_map(map_file_pointer, histogram, overwrite=True, dtype=np.int64)



import os

from hipscat.io.file_io import (
    append_paths_to_pointer,
    directory_has_contents,
    does_file_or_directory_exist,
    find_files_matching_path,
    get_directory_contents,
    get_file_pointer_from_path,
    is_regular_file,
)


def test_get_pointer_from_path(tmp_path):
    tmp_pointer = get_file_pointer_from_path(str(tmp_path))
    assert str(tmp_pointer) == str(tmp_path)


def test_file_or_dir_exist(small_sky_dir):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir)
    assert does_file_or_directory_exist(small_sky_pointer)
    catalog_info_string = os.path.join(small_sky_dir, "catalog_info.json")
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_string)
    assert does_file_or_directory_exist(catalog_info_pointer)


def test_file_or_dir_exist_false(small_sky_dir):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir + "incorrect file")
    assert not does_file_or_directory_exist(small_sky_pointer)


def test_append_paths_to_pointer(tmp_path):
    test_paths = ["folder", "file.txt"]
    test_path = os.path.join(tmp_path, *test_paths)
    tmp_pointer = get_file_pointer_from_path(str(tmp_path))
    assert append_paths_to_pointer(tmp_pointer, *test_paths) == test_path


def test_is_regular_file(small_sky_dir):
    partition_info_file = os.path.join(small_sky_dir, "partition_info.csv")
    assert is_regular_file(partition_info_file)

    assert not is_regular_file(small_sky_dir)

    partition_dir = os.path.join(small_sky_dir, "Norder=0")
    assert not is_regular_file(partition_dir)


def test_find_files_matching_path(small_sky_dir):
    ## no_wildcard
    assert len(find_files_matching_path(small_sky_dir, "partition_info.csv")) == 1

    ## wilcard in the name
    assert len(find_files_matching_path(small_sky_dir, "*.csv")) == 1


def test_find_files_matching_path_directory(small_sky_order1_dir):
    assert len(find_files_matching_path(small_sky_order1_dir)) == 1

    ## wildcard in directory - will match all files at indicated depth
    assert len(find_files_matching_path(small_sky_order1_dir, "*", "*", "*")) == 4


def test_directory_has_contents(small_sky_order1_dir, tmp_path):
    assert directory_has_contents(small_sky_order1_dir)
    assert not directory_has_contents(tmp_path)


def test_get_directory_contents(small_sky_order1_dir, tmp_path):
    small_sky_contents = get_directory_contents(small_sky_order1_dir)
    assert len(small_sky_contents) == 4

    expected = [
        os.path.join(small_sky_order1_dir, "Norder=1"),
        os.path.join(small_sky_order1_dir, "catalog_info.json"),
        os.path.join(small_sky_order1_dir, "partition_info.csv"),
        os.path.join(small_sky_order1_dir, "point_map.fits"),
    ]

    assert small_sky_contents == expected

    assert len(get_directory_contents(tmp_path)) == 0



"""Utilities to build bounding boxes around healpixels that include a neighor margin."""

import astropy.units as u
import astropy.wcs as world_coordinate_system
import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord
from regions import PixCoord, PolygonSkyRegion

from . import pixel_margins as pm


def get_margin_scale(pixel_order, margin_threshold=0.1):
    """Get the scale value need to expand the pixel bounding box to include the `margin_threshold`.

    Args:
        pixel_order (int): the order of the pixel to which we're calculating the margin region for.
        margin_threshold (float): the size of the border region in degrees.
    Returns:
        a float representing the scale factor.
    """
    if margin_threshold <= 0.0:
        raise ValueError("margin_threshold must be greater than 0.")

    resolution = hp.nside2resol(2**pixel_order, arcmin=True) / 60.0
    scale = 1 + ((2 * margin_threshold) / resolution)
    return scale


def get_margin_bounds_and_wcs(pixel_order, pix, scale, step=10):
    """Get the astropy `regions.PolygonPixelRegion` and `astropy.wcs` objects
        for a given margin bounding box scale.
        For healpixels that fall along the poles, the bounding boxes won't
        fully cover the expected margin area.

    Args:
        pixel_order (int): the order of the pixel to which we're calculating the margin region for.
        pix (int): the healpixel which we're calculating the margin region for.
        scale (float): the scale factor to apply to the given healpixel.
        step (int): the amount of samples of one side of the given healpixel's boundaries.
            total samples taken = 4 * step.
    Returns:
        polygons (list of tuples): a list of :obj:`regions.PolygonPixelRegion`
            and :obj:`wcs.WCS` tuples, covering the full area of the given healpixels
            scaled up area.
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    corners = hp.vec2dir(hp.boundaries(2**pixel_order, pix, step=1, nest=True), lonlat=True)

    min_ra = corners[0][1]  # western corner
    max_ra = corners[0][3]  # eastern corner
    min_dec = corners[1][2]  # southern corner
    max_dec = corners[1][0]  #  northern corner

    pixel_boundaries = hp.vec2dir(hp.boundaries(2**pixel_order, pix, step=step, nest=True), lonlat=True)

    # if the eastern corner is less than the western corner, then we've hit the
    # ra rollover and need to normalize to 0 -> 360.
    if max_ra < min_ra:
        max_ra = max_ra + 360.0
        ra_vals = pixel_boundaries[0]
        normal = np.where(ra_vals < 0.0, ra_vals + 360.0, ra_vals)
        pixel_boundaries[0] = normal

    # find the translation values to keep the bounding box
    # centered around the orignal healpixel.
    n_samples = len(pixel_boundaries[0])
    centroid_lon = np.sum(pixel_boundaries[0]) / n_samples
    centroid_lat = np.sum(pixel_boundaries[1]) / n_samples
    translate_lon = centroid_lon - (centroid_lon * scale)
    translate_lat = centroid_lat - (centroid_lat * scale)

    affine_matrix = np.array([[scale, 0, translate_lon], [0, scale, translate_lat], [0, 0, 1]])

    # convert the orignal boundary coordinates into
    # a homogenous coordinate space (3-dim)
    homogeneous = np.ones((3, n_samples))
    homogeneous[0] = pixel_boundaries[0]
    homogeneous[1] = pixel_boundaries[1]

    transformed_bounding_box = np.matmul(affine_matrix, homogeneous)

    # if the transform places the declination of any points outside of
    # the range 90 > dec > -90, change it to a proper dec value.
    transformed_bounding_box[1] = np.clip(transformed_bounding_box[1], -90.0, 90.0)

    # one arcsecond
    pix_size = 0.0002777777778

    ra_naxis = int((max_ra - min_ra) / pix_size)
    dec_naxis = int((max_dec - min_dec) / pix_size)

    polygons = []
    # for k < 2, the size of the bounding boxes breaks the regions
    # code, so we subdivide the pixel into multiple parts with
    # independent wcs.
    if pixel_order < 2:
        # k == 0 -> 16 subdivisions, k == 1 -> 4 subdivisions
        divs = 4 ** (2 - pixel_order)
        div_len = int((step * 4) / divs)
        for i in range(divs):
            j = i * div_len
            # get a `div_len + 2`  subset of the boundaries
            # inclusive of the first two elements of the next slice
            # so that we don't have any gaps in our boundary area.
            lon = list(transformed_bounding_box[0][j : j + div_len + 2])
            lat = list(transformed_bounding_box[1][j : j + div_len + 2])

            # in the last div, include the first two data points
            # to ensure that we cover the whole area.
            if i == divs - 1:
                lon.extend(list(transformed_bounding_box[0][0:2]))
                lat.extend(list(transformed_bounding_box[1][0:2]))

            # add the centroid as the last point, turning our set of
            # colinear-ish points into a slice of the total boundary box.
            lon.append(centroid_lon)
            lat.append(centroid_lat)

            ra_axis = int(ra_naxis / (2 ** (2 - pixel_order)))
            dec_axis = int(dec_naxis / (2 ** (2 - pixel_order)))

            wcs_partial = world_coordinate_system.WCS(naxis=2)
            wcs_partial.wcs.crpix = [1, 1]
            wcs_partial.wcs.crval = [lon[0], lat[0]]
            wcs_partial.wcs.cunit = ["deg", "deg"]
            wcs_partial.wcs.ctype = ["RA---TAN", "DEC--TAN"]
            wcs_partial.wcs.cdelt = [pix_size, pix_size]
            wcs_partial.array_shape = [ra_axis, dec_axis]

            vertices = SkyCoord(lon, lat, unit="deg")
            sky_region = PolygonSkyRegion(vertices=vertices)
            polygons.append((sky_region.to_pixel(wcs_partial), wcs_partial))
    else:
        wcs_margin = world_coordinate_system.WCS(naxis=2)
        wcs_margin.wcs.crpix = [1, 1]
        wcs_margin.wcs.crval = [min_ra, min_dec]
        wcs_margin.wcs.cunit = ["deg", "deg"]
        wcs_margin.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        wcs_margin.wcs.cdelt = [pix_size, pix_size]
        wcs_margin.array_shape = [ra_naxis, dec_naxis]

        vertices = SkyCoord(transformed_bounding_box[0], transformed_bounding_box[1], unit="deg")
        sky_region = PolygonSkyRegion(vertices=vertices)
        polygon_region = sky_region.to_pixel(wcs_margin)
        polygons = [(polygon_region, wcs_margin)]

    return polygons


def check_margin_bounds(r_asc, dec, poly_and_wcs):
    """Get the astropy `regions.PolygonPixelRegion` and `astropy.wcs` objects
        for a given margin bounding box scale.
        For pixels that fall along the poles, this code must be used in conjunction
        with `check_polar_margin_bounds` and
        `pixel_margins.get_truncated_margin_pixels`.

    Args:
        r_asc (:obj:`np.array`): one dimmensional array representing the ra of a
            given set of points.
        dec (:obj:`np.array`): one dimmensional array representing the dec of a
            given set of points.
        polygons (list of tuples): a list of obj:`regions.PolygonPixelRegion` and
            :obj:`wcs.WCS` tuples, covering the full area of the given healpixels
            scaled up area. see get_margin_bounds_and_wcs for more details.
    Returns:
        :obj:`numpy.array` of boolean values corresponding to the ra and dec
            coordinates checked against whether a given point is within any of the
            provided polygons.
    """
    sky_coords = SkyCoord(r_asc, dec, unit="deg")
    bound_vals = []
    for poly, wcs in poly_and_wcs:
        x_coords, y_coords = world_coordinate_system.utils.skycoord_to_pixel(sky_coords, wcs)
        pix_coords = PixCoord(x_coords, y_coords)
        vals = poly.contains(pix_coords)
        bound_vals.append(vals)
    return np.array(bound_vals).any(axis=0)


# pylint: disable=too-many-locals
def check_polar_margin_bounds(r_asc, dec, order, pix, margin_order, margin_threshold, step=1000):
    """Given a set of ra and dec values that are around one of the poles,
        determine if they are within `margin_threshold` of a provided
        partition pixel. This method helps us solve the edge cases that
        occur for margin bounding around the poles.

    Args:
        r_asc (:obj:`np.array`): one dimmensional array representing the ra of a
            given set of points.
        dec (:obj:`np.array`): one dimmensional array representing the dec of a
            given set of points.
        order (int): the healpix order of the provided partition pixel.
        pix (int): the partition healpixel that we are creating the margin cache of.
        margin_order (int): the healpix order of our pre-generated margin pixels.
        margin_threshold (float): the size of the border region in degrees.
        step (int): the total number of samples taken per side of the partition
            pixel `pix`. Only a small subset of these points will actually be used
            to check distance from the pixel, but step can be increased to get
            more granularity in the `margin_threshold` angular distance measurements.
    Returns:
        :obj:`numpy.array` of boolean values corresponding to the ra and dec
            coordinates checked against whether a given point is within any of the
            provided polygons.
    """
    polar, pole = pm.pixel_is_polar(order, pix)

    if not polar:
        raise ValueError("provided healpixel must be polar")

    part_pix_res = hp.nside2resol(2**order)
    marg_pix_res = hp.nside2resol(2**margin_order)

    # get the approximate number of boundary samples to cover a highest_k pixel
    # on the boundary of the main pixel
    boundary_range = int((marg_pix_res / part_pix_res) * step)
    pixel_boundaries = hp.vec2dir(hp.boundaries(2**order, pix, step=step, nest=True), lonlat=True)

    # to optimize our code, we only want to take boundary samples from the part
    # of the pixel that directly abuts the polar margin pixels.
    if pole == "North":
        end = len(pixel_boundaries[0])
        east_ra = pixel_boundaries[0][0 : boundary_range + 1]
        east_dec = pixel_boundaries[1][0 : boundary_range + 1]

        west_ra = pixel_boundaries[0][end - boundary_range : end]
        west_dec = pixel_boundaries[1][end - boundary_range : end]

        bound_ra = np.concatenate((east_ra, west_ra), axis=None)
        bound_dec = np.concatenate((east_dec, west_dec), axis=None)
        polar_boundaries = np.array([bound_ra, bound_dec])
    else:
        start = (2 * step) - boundary_range
        end = (2 * step) + boundary_range + 1
        south_ra = pixel_boundaries[0][start:end]
        south_dec = pixel_boundaries[1][start:end]
        polar_boundaries = np.array([south_ra, south_dec])

    # healpy.boundaries sometimes returns dec values greater than 90, especially
    # when taking many samples...
    polar_boundaries[1] = np.clip(polar_boundaries[1], -90.0, 90.0)

    sky_coords = SkyCoord(r_asc, dec, unit="deg")

    checks = []
    for i in range(len(polar_boundaries[0])):
        lon = polar_boundaries[0][i]
        lat = polar_boundaries[1][i]
        bound_coord = SkyCoord(lon, lat, unit="deg")

        ang_dist = bound_coord.separation(sky_coords)
        checks.append(ang_dist <= margin_threshold * u.deg)

    return np.array(checks).any(axis=0)



from __future__ import annotations

from dataclasses import dataclass

from hipscat.pixel_math.hipscat_id import HIPSCAT_ID_HEALPIX_ORDER


@dataclass(eq=True, frozen=True, order=True)
class HealpixPixel:
    """A HEALPix pixel, represented by an order and pixel number in NESTED ordering scheme

    see https://lambda.gsfc.nasa.gov/toolbox/pixelcoords.html for more information
    """

    order: int
    pixel: int

    def __post_init__(self) -> None:
        """Initialize a HEALPix pixel
        Args:
            order: HEALPix order
            pixel: HEALPix pixel number in NESTED ordering scheme
        """
        if self.order > HIPSCAT_ID_HEALPIX_ORDER:
            raise ValueError(f"HEALPix order cannot be greater than {HIPSCAT_ID_HEALPIX_ORDER}")

    def __str__(self) -> str:
        return f"Order: {self.order}, Pixel: {self.pixel}"

    def __repr__(self):
        return self.__str__()



from __future__ import annotations

from hipscat.pixel_math import HealpixInputTypes, get_healpix_pixel
from hipscat.pixel_tree.pixel_node import PixelNode


class PixelTree:
    """Sparse Quadtree of HEALPix pixels that make up the HiPSCat catalog

    This class stores each node in the tree, with leaf nodes corresponding to pixels with data
    files.

    There are a number of methods in this class which allow for quickly navigating through the
    tree and performing operations to filter the pixels in the catalog.

    Attributes:
        pixels: Nested dictionary of pixel nodes stored in the tree. Indexed by HEALPix
            order then pixel number
        root_pixel: Root node of the tree. Its children are a subset of the
            12 base HEALPix pixels
    """

    def __init__(self, root_pixel: PixelNode, pixels: dict[int, dict[int, PixelNode]]) -> None:
        """Initialises a tree object from the nodes in the tree

        Args:
            root_pixel: PixelNode representing the root node in the tree
            pixels: Dictionary containing all PixelNodes in the tree
        """
        self.root_pixel = root_pixel
        self.pixels = pixels

    def __len__(self):
        """Gets the number of nodes in the tree, including inner nodes and the root node

        Returns:
            The number of nodes in the tree, including inner nodes and the root node
        """
        pixel_count = 0
        for _, order_pixels in self.pixels.items():
            pixel_count += len(order_pixels)
        return pixel_count

    def contains(self, pixel: HealpixInputTypes) -> bool:
        """Check if tree contains a node at a given order and pixel

        Args:
            pixel: HEALPix pixel to check. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            True if the tree contains the pixel, False if not
        """
        pixel = get_healpix_pixel(pixel)
        return pixel.order in self.pixels and pixel.pixel in self.pixels[pixel.order]

    def __contains__(self, item):
        return self.contains(item)

    def get_node(self, pixel: HealpixInputTypes) -> PixelNode | None:
        """Get the node at a given pixel

        Args:
            pixel: HEALPix pixel to get. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            The PixelNode at the index, or None if a node does not exist
        """
        pixel = get_healpix_pixel(pixel)
        if self.contains(pixel):
            return self.pixels[pixel.order][pixel.pixel]
        return None

    def __getitem__(self, item):
        return self.get_node(item)
